from typing import Any, List, Literal

from ....utils.class_registry import register_class
from ...struct import Cost

from ...action import (
    ActionTypes,
    Actions,
    CreateObjectAction,
    DrawCardAction,
    MakeDamageAction,
    RemoveObjectAction,
)

from ...event import (
    MakeDamageEventArguments,
    RoundEndEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
)

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    IconType,
    ObjectPositionType,
    SkillType,
)

from ...modifiable_values import (
    CostValue,
    DamageElementEnhanceValue,
    DamageIncreaseValue,
    DamageValue,
)
from .base import (
    DefendCharacterStatus,
    ElementalInfusionCharacterStatus,
    PrepareCharacterStatus,
    ReviveCharacterStatus,
    RoundCharacterStatus,
    RoundEndAttackCharacterStatus,
    ShieldCharacterStatus,
    UsageCharacterStatus,
)


class Stealth_3_3(DefendCharacterStatus, ElementalInfusionCharacterStatus):
    name: Literal["Stealth"] = "Stealth"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    infused_elemental_type: DamageElementalType = DamageElementalType.PYRO

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self,
        value: DamageElementEnhanceValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageElementEnhanceValue:
        """
        When self use skill, and has talent, change physical to pyro
        """
        assert mode == "REAL"
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.talent is None:
            # no talent, do nothing
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(value, match, mode)

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        When self use skill, increase damage by 1
        """
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not corresponding character use damage skill, do nothing
            return value
        if self.usage <= 0:  # pragma: no cover
            # no usage, do nothing
            return value
        # increase damage by 1
        value.damage += 1
        self.usage -= 1
        return value


class ExplosiveSpark_3_4(UsageCharacterStatus):
    name: Literal["Explosive Spark"] = "Explosive Spark"
    version: Literal["3.4"] = "3.4"
    usage: int
    max_usage: int
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this character use normal attack, not modify
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack, not modify
            return value
        # modify damage
        assert self.usage > 0
        self.usage -= 1
        value.damage += 1
        return value

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        If self use charged attack, decrease one
        unaligned die cost.
        """
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not this character use skill, not modify
            return value
        skill = match.get_object(value.position)
        skill_type: SkillType = skill.skill_type
        if skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack, not modify
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack, not modify
            return value
        # try decrease pyro cost
        value.cost.decrease_cost(DieColor.PYRO)
        return value


class NiwabiEnshou_3_3(ElementalInfusionCharacterStatus, UsageCharacterStatus):
    name: Literal["Niwabi Enshou"] = "Niwabi Enshou"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    infused_elemental_type: DamageElementalType = DamageElementalType.PYRO
    icon_type: Literal[IconType.ATK_UP_FIRE] = IconType.ATK_UP_FIRE

    effect_triggered: bool = False

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        If corresponding character use normal attack, increase damage by 1,
        and mark effect triggered.
        """
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this character use normal attack, not modify
            return value
        # this character use normal attack, modify
        self.usage -= 1
        self.effect_triggered = True
        value.damage += 1
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        do not remove immediately here, as may trigger additional attack at
        skill end.
        """
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        If effect triggered, and have talent on yoimiya, make 1 pyro damage.
        If usage is zero, then remove self.
        """
        ret: List[MakeDamageAction | RemoveObjectAction] = []
        if self.position.player_idx != event.action.position.player_idx:
            # not self player use skill
            self.effect_triggered = False
            return list(self.check_should_remove())
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if (
            character.name == "Yoimiya"
            and event.action.skill_type == SkillType.NORMAL_ATTACK
            and self.effect_triggered
            and character.talent is not None
        ):
            # yoimiya use normal attack and self effect triggered
            # and has talent, attack 1 pyro damage
            target = match.player_tables[
                1 - self.position.player_idx
            ].get_active_character()
            ret.append(
                MakeDamageAction(
                    damage_value_list=[
                        DamageValue(
                            position=self.position,
                            damage_type=DamageType.DAMAGE,
                            target_position=target.position,
                            damage=1,
                            damage_elemental_type=DamageElementalType.PYRO,
                            cost=Cost(),
                        )
                    ]
                )
            )
        # reset mark
        self.effect_triggered = False
        return ret + self.check_should_remove()


class Brilliance_3_8(RoundCharacterStatus):
    name: Literal["Brilliance"] = "Brilliance"
    version: Literal["3.8"] = "3.8"
    usage: int = 2
    max_usage: int = 2

    decrease_cost_usage: int = 1
    decrease_cost_max_usage: int = 1
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def renew(self, new_status: "Brilliance_3_8") -> None:
        self.decrease_cost_usage = max(
            new_status.decrease_cost_usage, self.decrease_cost_usage
        )
        self.decrease_cost_max_usage = max(
            new_status.decrease_cost_max_usage, self.decrease_cost_max_usage
        )
        return super().renew(new_status)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.decrease_cost_usage = self.decrease_cost_max_usage
        return super().event_handler_ROUND_PREPARE(event, match)

    def value_modifier_COST(
        self,
        value: CostValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> CostValue:
        if self.decrease_cost_usage <= 0:
            # no usage left
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charge attack
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not self use skill
            return value
        skill = match.get_object(value.position)
        if skill.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack
            return value
        # decrease cost
        if value.cost.decrease_cost(DieColor.PYRO):  # pragma: no branch
            if mode == "REAL":
                self.decrease_cost_usage -= 1
        return value

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        return [
            CreateObjectAction(
                object_name="Scarlet Seal",
                object_position=self.position,
                object_arguments={},
            )
        ]


class ScarletSeal_4_2(UsageCharacterStatus):
    name: Literal["Scarlet Seal"] = "Scarlet Seal"
    version: Literal["4.2"] = "4.2"
    usage: int = 1
    max_usage: int = 2
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP
    triggered: bool = False

    available_handler_in_trashbin: List[ActionTypes] = [ActionTypes.SKILL_END]

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not self use normal attack
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charge attack
            return value
        # increase damage
        assert mode == "REAL"
        self.usage -= 1
        self.triggered = True
        value.damage += 2
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[DrawCardAction]:
        """
        When triggered, and self character have talent, talent need to draw,
        then draw one card.
        """
        if not self.triggered:
            # not triggered
            return []
        self.triggered = False
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.talent is not None and character.talent.draw_card:
            # have talent and need draw
            return [
                DrawCardAction(
                    player_idx=self.position.player_idx,
                    number=1,
                    draw_if_filtered_not_enough=True,
                )
            ]
        return []


class ParamitaPapilio_3_7(ElementalInfusionCharacterStatus, RoundCharacterStatus):
    name: Literal["Paramita Papilio"] = "Paramita Papilio"
    version: Literal["3.7"] = "3.7"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    infused_elemental_type: DamageElementalType = DamageElementalType.PYRO

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not self use damage skill
            return value
        if value.damage_elemental_type != DamageElementalType.PYRO:  # pragma: no cover
            # not pyro damage
            return value
        # increase damage
        value.damage += 1
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If self use charged attack, create blood blossom on target
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not self use skill
            return []
        if not (
            event.action.skill_type == SkillType.NORMAL_ATTACK
            and match.player_tables[self.position.player_idx].charge_satisfied
        ):
            # not charged attack
            return []
        return [
            CreateObjectAction(
                object_position=event.action.target_position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_name="Blood Blossom",
                object_arguments={},
            )
        ]


class BloodBlossom_3_7(RoundEndAttackCharacterStatus):
    name: Literal["Blood Blossom"] = "Blood Blossom"
    version: Literal["3.7"] = "3.7"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.DOT] = IconType.DOT
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO


class DilucInfusion_3_3(ElementalInfusionCharacterStatus, RoundCharacterStatus):
    name: Literal["Pyro Elemental Infusion"] = "Pyro Elemental Infusion"
    mark: Literal["Diluc"] = "Diluc"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2


class FieryRebirth_3_7(ReviveCharacterStatus):
    name: Literal["Fiery Rebirth"] = "Fiery Rebirth"
    version: Literal["3.7"] = "3.7"
    heal: int = 3


class AegisOfAbyssalFlame_3_7(ShieldCharacterStatus):
    name: Literal["Aegis of Abyssal Flame"] = "Aegis of Abyssal Flame"
    version: Literal["3.7"] = "3.7"
    usage: int = 3
    max_usage: int = 3

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        if (
            value.is_corresponding_character_use_damage_skill(
                self.position, match, None
            )
            and value.damage_elemental_type == DamageElementalType.PYRO
        ):
            # self attack and pyro damage, increase 1 damage
            assert mode == "REAL"
            value.damage += 1
        return value


class IncinerationDrive_4_1(PrepareCharacterStatus):
    name: Literal["Incineration Drive"] = "Incineration Drive"
    version: Literal["4.1"] = "4.1"
    character_name: Literal["Dehya"] = "Dehya"
    skill_name: Literal["Incineration Drive"] = "Incineration Drive"


class ScarletSeal_3_8(ScarletSeal_4_2):
    version: Literal["3.8"]
    usage: int = 1
    max_usage: int = 1


register_class(
    Stealth_3_3
    | ExplosiveSpark_3_4
    | NiwabiEnshou_3_3
    | Brilliance_3_8
    | ScarletSeal_4_2
    | ParamitaPapilio_3_7
    | BloodBlossom_3_7
    | DilucInfusion_3_3
    | FieryRebirth_3_7
    | AegisOfAbyssalFlame_3_7
    | IncinerationDrive_4_1
    | ScarletSeal_3_8
)



from lpsim.agents.interaction_agent import InteractionAgent_V1_0, InteractionAgent
from lpsim.agents.nothing_agent import NothingAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    get_pidx_cidx,
    get_test_id_from_command,
    make_respond,
    get_random_state,
    set_16_omni,
)
from lpsim.server.interaction import UseSkillRequest


def test_small_elemental_artifacts():
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent_V1_0(
        version="1.0",
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "reroll 1 2 3 4 5 6 7",
            "card 4 0 pyro geo",
            # cost decrease
            "skill 0 dendro hydro",
            # check skill cost after first attack (should no cost decrease)
            "tune hydro 1",
            "card 1 0 dendro dendro",
            # after equip second artifact, can use A but not E
            # (cost is dendro + any)
            "skill 0 omni omni",
            "end",
            "reroll 0 1 2 3 4 5 6 7",
            "card 0 0 anemo",
            "tune hydro 0",
            # next run after tune, 3 skills are available but only normal
            # any decrease
            "skill 1 dendro dendro omni",
            "tune pyro 0",
            # after skill 1 and tune, still have 1-1 normal attack
            "card 0 0 electro electro",
            "tune hydro 0",
            "skill 2 dendro dendro",
            "end",
            "reroll",
            "sw_char 1 anemo",
            "sw_char 2 anemo",
            "tune dendro 0",
            "skill 0 electro dendro dendro",
            "tune pyro 0",
            "skill 0 electro cryo cryo",
            "end",
            # 85 10 10 99 10 10
        ],
        only_use_command=True,
    )
    match = Match(version="0.0.1", random_state=get_random_state())
    deck = {
        "name": "Deck",
        "characters": [
            {
                "name": "DendroMobMage",
                "element": "DENDRO",
                "hp": 99,
                "max_hp": 99,
            },
            {
                "name": "DendroMobMage",
                "element": "DENDRO",
            },
            {
                "name": "ElectroMobMage",
                "element": "ELECTRO",
            },
        ],
        "cards": [
            {
                "name": "Wine-Stained Tricorne",
            }
        ]
        * 10
        + [
            {
                "name": "Laurel Coronet",
            }
        ]
        * 10
        + [
            {
                "name": "Strategize",
            }
        ]
        * 10,
    }
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            # asserts
            if len(agent_1.commands) == 22:
                # cost decrease
                skills = [x for x in match.requests if x.name == "UseSkillRequest"]
                assert len(skills) == 2
                assert skills[0].cost.elemental_dice_number == 0
                assert skills[0].cost.any_dice_number == 2
                assert skills[1].cost.elemental_dice_number == 2
                assert skills[1].cost.any_dice_number == 0
            elif len(agent_1.commands) == 21:
                # check skill cost after first attack (should no cost decrease)
                skills = [x for x in match.requests if x.name == "UseSkillRequest"]
                assert len(skills) == 2
                assert skills[0].cost.elemental_dice_number == 1
                assert skills[0].cost.any_dice_number == 2
                assert skills[1].cost.elemental_dice_number == 3
                assert skills[1].cost.any_dice_number == 0
            elif len(agent_1.commands) == 19:
                # after equip second artifact, can use A but not E
                # (cost is dendro + any)
                skills = [x for x in match.requests if x.name == "UseSkillRequest"]
                assert len(skills) == 1
                skill: UseSkillRequest = skills[0]
                assert skill.cost.elemental_dice_number == 1
                assert skill.cost.any_dice_number == 1
            elif len(agent_1.commands) == 14:
                # next run after tune, 3 skills are available but only normal
                # any decrease
                skills = [x for x in match.requests if x.name == "UseSkillRequest"]
                assert len(skills) == 3
                skills.sort(key=lambda x: x.skill_idx)
                assert skills[0].cost.elemental_dice_number == 1
                assert skills[0].cost.any_dice_number == 1
                assert skills[1].cost.elemental_dice_number == 3
                assert skills[1].cost.any_dice_number == 0
                assert skills[2].cost.elemental_dice_number == 3
                assert skills[2].cost.any_dice_number == 0
            elif len(agent_1.commands) == 12:
                # after skill 1 and tune, still have 1-1 normal attack
                skills = [x for x in match.requests if x.name == "UseSkillRequest"]
                assert len(skills) == 1
                skill: UseSkillRequest = skills[0]
                assert skill.cost.elemental_dice_number == 1
                assert skill.cost.any_dice_number == 1

            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 4
    check_hp(match, [[85, 10, 10], [99, 10, 10]])
    assert match.player_tables[1].characters[0].artifact is not None

    assert match.state != MatchState.ERROR


def test_create_small_element_artifacts():
    deck = Deck.from_str(
        """
        character:Nahida*3
        Broken Rime's Echo*4
        Laurel Coronet*4
        Mask of Solitude Basalt*4
        Thunder Summoner's Crown*4
        Viridescent Venerer's Diadem*4
        Wine-Stained Tricorne*4
        Witch's Scorching Hat*4
        Strategize*2
        """
    )
    match = Match()
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 4
    match.config.max_hand_size = 30
    match.config.initial_hand_size = 30
    match.start()
    match.step()

    assert match.state != MatchState.ERROR


def test_old_version_elemental_artifacts():
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "reroll 1 4",
            "TEST 1 all card can use",
            "sw_char 1 4",
            "TEST 1 all card can use",
            "sw_char 2 0",
            "TEST 2 3 card can use, 0 and 4 cannot",
            "end",
        ],
        only_use_command=True,
    )
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Mona
        character:Nahida
        Wine-Stained Tricorne*12
        Timmie*2
        Rana*2
        Strategize*2
        # Timmie*22
        """
    )
    old_wine = {"name": "Wine-Stained Tricorne", "version": "3.3"}
    deck_dict = deck.dict()
    deck_dict["cards"] += [old_wine] * 12
    deck = Deck(**deck_dict)
    match = Match(random_state=get_random_state())
    match.config.max_same_card_number = 30
    match.set_deck([deck, deck])
    match.start()
    match.step()
    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    counter = 0
                    for request in match.requests:
                        if request.name == "UseCardRequest":
                            counter += 1
                    assert counter == 5
                elif test_id == 2:
                    counter = []
                    for request in match.requests:
                        if request.name == "UseCardRequest":
                            counter.append(request.card_idx)
                    assert len(counter) == 3
                    assert 0 not in counter
                    assert 4 not in counter
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0

    assert match.state != MatchState.ERROR


def test_gambler():
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "end",
            "choose 1",
            "choose 2",
            "choose 3",
            "choose 4",
            "choose 5",
            "choose 6",
            "skill 0 0 1 2",
            "choose 7",
            "card 0 0 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "TEST 1 9 15",
            "end",
            "skill 0 0 1 2",
            "TEST 1 13 15",
            "card 0 0 0",
            "skill 0 0 1 2",
            "TEST 1 11 15",
            "end",
            "TEST 2 p0c7 usage2",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "card 0 1 1",
            "skill 0 0 1 2",
            "TEST 1 16 12",
            "card 0 0 1",
            "skill 0 0 1 2",
            "TEST 1 16 10",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "TEST 1 16 8",
            "skill 0 0 1 2",
            "TEST 1 16 5",
            "card 0 0 0",
            "skill 0 0 1 2",
            "TEST 1 16 3",
            "card 0 2 0",
            "end",
            "choose 1",
            "skill 0 0 1 2",
            "choose 2",
            "TEST 1 11 15",
            "end",
            "choose 3",
            "choose 4",
            "card 0 0 0",
            "end",
            "choose 5",
            "choose 6",
            "end",
        ],
        only_use_command=True,
    )
    deck = Deck.from_str(
        """
        # character:Fischl
        # character:Mona
        # character:Nahida
        character:Nahida*10
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Timmie*2
        # Rana*2
        # Strategize*2
        # The Bestest Travel Companion!*2
        # Covenant of Rock
        Gambler's Earrings*30
        """
    )
    # old_wine = {'name': 'Wine-Stained Tricorne', 'version': '3.3'}
    # deck_dict = deck.dict()
    # deck_dict['cards'] += [old_wine] * 12
    # deck = Deck(**deck_dict)
    for character in deck.characters:
        character.hp = 1
        character.max_hp = 1
    match = Match(random_state=get_random_state())
    match.config.max_same_card_number = 30
    match.config.character_number = 10
    match.config.random_first_player = False
    match.set_deck([deck, deck])
    set_16_omni(match)
    match.start()
    match.step()
    while True:
        if match.need_respond(0):
            while True:
                cmd = agent_0.commands[0]
                test_id = get_test_id_from_command(agent_0)
                if test_id == 1:
                    number = [int(x) for x in cmd.strip().split(" ")[-2:]]
                    assert number[0] == len(match.player_tables[0].dice.colors)
                    assert number[1] == len(match.player_tables[1].dice.colors)
                elif test_id == 2:
                    character = match.player_tables[0].characters[7]
                    assert character.artifact is not None
                    assert character.artifact.name == "Gambler's Earrings"
                else:
                    break
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                cmd = agent_1.commands[0]
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    number = [int(x) for x in cmd.strip().split(" ")[-2:]]
                    assert number[0] == len(match.player_tables[0].dice.colors)
                    assert number[1] == len(match.player_tables[1].dice.colors)
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_1.commands) == 0 and len(agent_0.commands) == 0

    assert match.state != MatchState.ERROR


def test_old_gambler():
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "end",
            "choose 1",
            "choose 2",
            "choose 3",
            "choose 4",
            "choose 5",
            "choose 6",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "card 0 1 1",
            "skill 0 0 1 2",
            "TEST 1 16 12",
            "card 0 0 1",
            "skill 0 0 1 2",
            "TEST 1 16 10",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "TEST 1 16 8",
            "skill 0 0 1 2",
            "TEST 1 16 7",
            "skill 0 0 1 2",
            "TEST 1 16 6",
            "end",
        ],
        only_use_command=True,
    )
    deck = Deck.from_str(
        """
        # character:Fischl
        # character:Mona
        # character:Nahida
        character:Nahida*10
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Timmie*2
        # Rana*2
        # Strategize*2
        # The Bestest Travel Companion!*2
        # Covenant of Rock
        # Gambler's Earrings*30
        """
    )
    old_gambler = {"name": "Gambler's Earrings", "version": "3.3"}
    deck_dict = deck.dict()
    deck_dict["cards"] += [old_gambler] * 30
    deck = Deck(**deck_dict)
    for character in deck.characters:
        character.hp = 1
        character.max_hp = 1
    match = Match(random_state=get_random_state())
    match.config.max_same_card_number = 30
    match.config.character_number = 10
    match.config.random_first_player = False
    match.set_deck([deck, deck])
    set_16_omni(match)
    match.start()
    match.step()
    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                cmd = agent_1.commands[0]
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    number = [int(x) for x in cmd.strip().split(" ")[-2:]]
                    assert number[0] == len(match.player_tables[0].dice.colors)
                    assert number[1] == len(match.player_tables[1].dice.colors)
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_1.commands) == 0 and len(agent_0.commands) == 0

    assert match.state != MatchState.ERROR


def test_millelith():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 2 0 0 1 2",
            "TEST 3 p0c0 no status",
            "skill 0 0 1 2",
            "TEST 2 dice number 11 13",
            "TEST 1 9 10 10 9 10 10",
            "end",
            "TEST 4 p0c0 p1c1 status",
            "sw_char 1 0",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "TEST 1 7 10 10 9 10 10",
            "TEST 2 dice number 11 11",
            "sw_char 0 0",
            "TEST 2 dice number 11 8",
            "end",
            "TEST 4 p0c0 p1c1 status 2 usage",
            "end",
            "TEST 4 p0c0 p1c1 status 2 usage",
            "end",
        ],
        [
            "sw_card 0 1 2",
            "choose 0",
            "TEST 2 dice number 10 16",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "TEST 2 dice number 11 10",
            "TEST 1 8 10 10 9 10 10",
            "sw_char 1 0",
            "card 2 1 0 1",
            "end",
            "sw_char 0 0",
            "sw_char 1 0",
            "TEST 5 p1c1 status 1 usage 1",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "end",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:PyroMobMage
        character:ElectroMobMage
        character:Noelle
        Tenacity of the Millelith*15
        General's Ancient Helm*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.strip().split()
                hps = [int(x) for x in cmd[2:]]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.strip().split()
                number = [int(x) for x in cmd[-2:]]
                assert number[0] == len(match.player_tables[0].dice.colors)
                assert number[1] == len(match.player_tables[1].dice.colors)
            elif test_id == 3:
                assert len(match.player_tables[0].characters[0].status) == 0
            elif test_id == 4:
                assert len(match.player_tables[0].characters[0].status) == 1
                assert match.player_tables[0].characters[0].status[0].usage == 2
                assert len(match.player_tables[1].characters[1].status) == 1
                assert match.player_tables[1].characters[1].status[0].usage == 2
            elif test_id == 5:
                assert len(match.player_tables[1].characters[1].status) == 1
                assert match.player_tables[1].characters[1].status[0].usage == 1
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_instructor():
    cmd_records = [
        [
            "sw_card 2 1 3",
            "choose 1",
            "card 1 0 15 14",
            "card 1 1 13 12",
            "card 1 2 11 10",
            "end",
            "TEST 2 p0 deck 3",
            "TEST 2 p1 deck 3",
            "end",
            "sw_char 0 15",
            "sw_char 2 14",
            "skill 0 13 12 11",
            "sw_char 0 10",
            "end",
            "end",
            "sw_char 2 15",
            "skill 0 14 13 12",
            "skill 0 11 10 9",
            "sw_char 0 8",
            "end",
            "skill 0 15 14 13",
            "skill 0 13 12 11",
            "skill 0 11 10 9",
            "skill 0 9 8 7",
            "card 0 0 6 5",
            "skill 0 4 3 2",
            "skill 2 2 1 0",
        ],
        [
            "sw_card 1 2",
            "choose 1",
            "skill 0 15 14 13",
            "sw_char 0 12",
            "skill 0 11 10 9",
            "sw_char 2 8",
            "skill 0 7 6 5",
            "TEST 3 p0 dice 10",
            "TEST 3 p1 dice 5",
            "sw_char 0 4",
            "end",
            "card 3 0 15 14",
            "skill 0 13 12 11",
            "TEST 3 p1 dice 12",
            "TEST 1 6 4 6 10 10 10",
            "sw_char 1 11",
            "skill 0 10 9 8",
            "sw_char 2 7",
            "skill 0 6 5 4",
            "TEST 3 p1 dice 4",
            "end",
            "end",
            "end",
            "sw_char 1 15",
            "sw_char 0 14",
            "sw_char 1 13",
            "end",
            "sw_char 2 15",
            "sw_char 1 14",
            "sw_char 2 13",
            "TEST 3 p0 dice 10",
            "sw_char 1 12",
            "TEST 3 p0 dice 7",
            "sw_char 2 11",
            "TEST 3 p0 dice 3",
            "sw_char 1 10",
            "choose 2",
            "TEST 3 p0 dice 1",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:AnemoMobMage
        character:Barbara
        character:Yae Miko
        Instructor's Cap*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.strip().split()
                hps = [int(x) for x in cmd[2:]]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.strip().split()
                pidx = int(cmd[2][1])
                num = int(cmd[-1])
                assert len(match.player_tables[pidx].table_deck) == num
            elif test_id == 3:
                cmd = cmd.strip().split()
                pidx = int(cmd[2][1])
                num = int(cmd[-1])
                assert len(match.player_tables[pidx].dice.colors) == num
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_lucky_dog():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "skill 0 15 14 13",
            "TEST 1 10 10 10 10 10 10",
            "skill 0 12 11 10",
            "card 0 1 9 8",
            "skill 0 7 6 5",
            "sw_char 0 4",
            "sw_char 1 3",
            "skill 3 2 1 0",
            "end",
            "skill 2 15 14 13 12 11",
            "TEST 1 10 3 10 6 5 6",
            "sw_char 0 10",
            "card 2 0 9 8",
            "skill 1 7 6 5",
            "skill 1 4 3 2",
        ],
        [
            "sw_card",
            "choose 0",
            "card 0 0 15 14",
            "skill 2 13 12 11 10 9",
            "skill 1 8 7 6",
            "sw_char 2 5",
            "sw_char 1 4",
            "skill 0 3 2 1",
            "TEST 1 10 8 10 7 8 9",
            "end",
            "sw_char 0 15",
            "TEST 1 10 3 10 4 5 6",
            "skill 1 14 13 12",
            "sw_char 2 11",
            "skill 0 10 9 8",
            "TEST 1 9 3 10 6 5 6",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Rhodeia of Loch
        character:Ganyu
        character:Yae Miko
        Lucky Dog's Silver Circlet*20
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.strip().split()
                hps = [int(x) for x in cmd[2:]]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_big_elemental_artifacts():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "reroll 4 3",
            "card 1 2 2 1 0",
            "card 1 1 0 1 2",
            "card 1 0 1 0",
            "end",
            "TEST 1 p0 pyro dendro geo",
            "reroll",
            "card 3 1 5 4 3",
            "sw_char 1 4",
            "TEST 2 skill 0 cost 2",
            "skill 0 0 3",
            "end",
            "TEST 1 p0 dendro geo geo",
            "reroll",
            "end",
        ],
        [
            "sw_card",
            "choose 2",
            "reroll 1 0 4 5",
            "card 2 2 7 6 5",
            "card 2 1 4 3 1",
            "end",
            "TEST 1 p1 electro electro",
            "reroll",
            "card 0 0 7 6 5",
            "end",
            "TEST 1 p1 cryo electro electro",
            "reroll",
            "end",
        ],
    ]
    repeat_times = 20
    for _ in range(repeat_times):
        cmd_records[0] += cmd_records[0][-2:]
        cmd_records[1] += cmd_records[1][-2:]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Xingqiu
        character:Kamisato Ayaka
        character:Yae Miko
        Blizzard Strayer*2
        Deepwood Memories*2
        Archaic Petra*2
        Thundering Fury*2
        Viridescent Venerer*2
        Heart of Depth*2
        Crimson Witch of Flames*2
        Blizzard Strayer@3.6*2
        Deepwood Memories@3.6*2
        Archaic Petra@3.6*2
        Thundering Fury@3.6*2
        Viridescent Venerer@3.6*2
        Heart of Depth@3.6*2
        Crimson Witch of Flames@3.6*2
        Blizzard Strayer@3.3*2
        Deepwood Memories@3.3*2
        Archaic Petra@3.3*2
        Thundering Fury@3.3*2
        Viridescent Venerer@3.3*2
        Heart of Depth@3.3*2
        Crimson Witch of Flames@3.3*2
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.config.max_round_number = 999
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                colors = [x.upper() for x in cmd[3:]]
                colors += colors
                for color in match.player_tables[pidx].dice.colors:
                    if color in colors:
                        colors.remove(color)
                assert len(colors) == 0
            elif test_id == 2:
                cmd = cmd.split()
                sid = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseSkillRequest" and req.skill_idx == sid:
                        assert req.cost.total_dice_cost == cost
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_advanturer_traveling_doctor():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 0 15 14 13",
            "card 12 0 12",
            "skill 0 11 10 9",
            "skill 0 8 7 6",
            "skill 0 5 4 3",
            "skill 0 2 1 0",
            "end",
            "skill 0 15 14 13",
            "skill 1 12 11 10",
            "TEST 1 24 30 30 27 29 24",
            "skill 2 9 8 7",
            "TEST 1 23 30 30 26 28 22",
            "sw_char 1",
            "skill 1 6 5 4",
            "sw_char 0",
            "skill 0 3 2 1",
            "end",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "skill 0 9 8 7",
            "skill 0 6 5 4",
            "skill 0 3 2 1",
            "end",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "end",
            "skill 1 15 14 13",
            "skill 0 12 11 10",
            "TEST 1 25 30 30 0 28 19",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "skill 0 15 14 13",
            "sw_char 1 12",
            "sw_char 2 11",
            "skill 0 10 9 8",
            "TEST 1 28 30 30 26 28 26",
            "sw_char 1 7",
            "skill 0 6 5 4",
            "card 5 2 3",
            "sw_char 2 2",
            "card 9 1",
            "card 1 0",
            "card 3 0",
            "end",
            "skill 1 15 14 13",
            "TEST 1 25 30 30 26 28 23",
            "skill 2 12 11 10",
            "TEST 1 24 30 30 26 28 22",
            "card 13 0 9",
            "skill 2 8 7 6",
            "sw_char 0 5",
            "end",
            "end",
            "end",
            "choose 2",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "skill 2 9 8 7",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Venti
        character:Xiangling
        character:Xingqiu
        Sweet Madame*15
        Adventurer's Bandana*5
        Traveling Doctor's Handkerchief*5
        Calx's Arts*5
        """
    )
    for character in deck.characters:
        character.max_hp = 30
        character.hp = 30
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.initial_hand_size = 20
    match.config.max_hand_size = 30
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_emblem_of_severed_fate():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 2 2 15 14",
            "card 10 1 13 12 11",
            "card 10 0 10",
            "skill 0 9 8 7",
            "skill 0 6 5 4",
            "skill 2 3 2 1",
            "end",
            "sw_char 1 15",
            "skill 0 14 13 12",
            "sw_char 0 11",
            "skill 0 10 9 8",
            "sw_char 1 7",
            "end",
            "skill 2 15 14 13 12",
            "card 9 0 11",
            "skill 2 10 9 8 7",
            "TEST 1 19 21 30 30 15 21",
            "sw_char 2 6",
            "TEST 1 19 21 29 30 15 21",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "card 0 0 15 14",
            "card 2 2 13 12",
            "skill 0 11 10 9",
            "skill 0 8 7 6",
            "TEST 1 26 30 30 30 24 30",
            "TEST 2 p0c0 charge 0",
            "TEST 2 p0c1 charge 1",
            "TEST 2 p0c2 charge 1",
            "TEST 2 p1c0 charge 0",
            "TEST 2 p1c1 charge 2",
            "TEST 2 p1c2 charge 0",
            "skill 2 5 4 3 2",
            "end",
            "sw_char 2 15",
            "skill 0 14 13 12",
            "sw_char 1 11",
            "skill 0 10 9 8",
            "sw_char 2 7",
            "end",
            "TEST 1 19 26 30 30 15 26",
            "sw_char 2 15",
            "TEST 1 19 26 30 30 15 21",
            "skill 2 14 13 12",
            "card 3 0 11",
            "skill 2 10 9 8",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Venti
        character:Xiangling
        character:Xingqiu@4.0
        Sweet Madame*15
        Ornate Kabuto*5
        Ornate Kabuto@3.7*5
        Emblem of Severed Fate*5
        Emblem of Severed Fate@4.0*5
        Emblem of Severed Fate@3.7*5
        Calx's Arts*5
        """
    )
    for character in deck.characters:
        character.max_hp = 30
        character.hp = 30
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.initial_hand_size = 20
    match.config.max_hand_size = 30
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                pidx, cidx = get_pidx_cidx(cmd)
                charge = int(cmd[-1])
                assert match.player_tables[pidx].characters[cidx].charge == charge
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_shadow_of_chiwang():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "sw_char 2 15",
            "sw_char 1 14",
            "sw_char 0 13",
            "card 0 0 12",
            "skill 1 11 10 9",
            "sw_char 1 8",
            "sw_char 2 7",
            "sw_char 1 6",
            "card 1 1 5",
            "skill 1 4 3 2",
            "TEST 1 hands 6 5",
            "end",
            "sw_char 2 15",
            "skill 1 14 13 12",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "sw_char 1 15",
            "card 1 1 14",
            "skill 1 13 12 11",
            "TEST 1 hands 5 5",
            "sw_char 2 10",
            "skill 1 9 8 7",
            "sw_char 1 6",
            "end",
            "TEST 1 hands 8 8",
            "skill 1 15 14 13",
            "sw_char 2 12",
            "skill 1 11 10 9",
            "sw_char 1 8",
            "skill 1 7 6 5",
            "TEST 1 hand 8 9",
            "sw_char 2 4",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Xingqiu
        character:Fischl
        character:Mona
        Shadow of the Sand King*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hlen = [int(cmd[3]), int(cmd[4])]
                assert len(match.player_tables[0].hands) == hlen[0]
                assert len(match.player_tables[1].hands) == hlen[1]
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_millelith_2():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "skill 1 9 8 7",
            "card 4 2 6 5 4",
            "end",
            "choose 1",
        ],
        [
            "sw_card",
            "choose 0",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "skill 1 9 8 7",
            "skill 0 6 5 4",
            "TEST 1 p0 dice 4",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Keqing@4.2*3
        Tenacity of the Millelith*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                assert len(match.player_tables[0].dice.colors) == 4
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # from tests.utils_for_test import enable_logging
    # enable_logging()
    # test_small_elemental_artifacts()
    # test_create_small_element_artifacts()
    # test_old_version_elemental_artifacts()
    # test_gambler()
    # test_old_gambler()
    # test_millelith()
    # test_instructor()
    # test_lucky_dog()
    # test_big_elemental_artifacts()
    # test_advanturer_traveling_doctor()
    test_emblem_of_severed_fate()
    # test_shadow_of_chiwang()
    test_millelith_2()



from ....utils import import_all_modules
from .base import SupportBase
from .companions import CompanionBase
from .items import ItemBase
from .locations import LocationBase


import_all_modules(__file__, __name__)
__all__ = ("SupportBase", "CompanionBase", "ItemBase", "LocationBase")



# tests that test corner cases to improve coverage

from lpsim.agents import InteractionAgent
from lpsim import Deck, Match, MatchState
from tests.utils_for_test import (
    check_hp,
    check_usage,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_lynette_single():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 2 0 15 14 13",
            "skill 1 12 11 10",
            "TEST 1",
            "end",
        ],
        ["sw_card", "choose 0", "skill 1 15 14 13", "skill 1 12 11 10"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Lynette
        A Cold Blade Like a Shadow*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                assert match.player_tables[0].characters[0].hp == 4
                assert match.player_tables[1].characters[0].hp == 4
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_gorou_alhaitham_layla():
    cmd_records = [
        [
            "sw_card 1 2",
            "choose 2",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "card 3 0",
            "skill 2 9 8 7",
            "end",
            "choose 1",
            "end",
            "choose 0",
            "end",
            "end",
        ],
        [
            "sw_card 2 3",
            "choose 1",
            "card 1 0 15 14 13",
            "skill 0 12 11 10",
            "skill 2 9 8 7",
            "sw_char 0 6",
            "skill 0 5 4 3",
            "skill 0 2 1 0",
            "end",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "skill 2 9 8 7",
            "end",
            "end",
            "skill 1 15 14 13",
            "sw_char 1 12",
            "skill 1 11 10 9",
            "TEST 1",
            "TEST 2",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Layla
        character:Gorou
        character:Alhaitham
        Rushing Hound: Swift as the Wind*10
        Sweet Madame*10
        Changing Shifts*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                for s in match.player_tables[1].team_status:
                    assert s.usage == 2
            elif test_id == 2:
                check_hp(match, [[5, 0, 0], [10, 2, 10]])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_dvalin_2():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "skill 1 15 14 13",
            "sw_char 0 12",
            "skill 1 11 10 9",
            "TEST 1 10 10 10 6 10 10",
            "skill 0 8 7 6",
            "TEST 1 10 10 10 2 10 10",
            "sw_char 1 5",
            "skill 0 4 3 2",
            "end",
            "skill 5 15 14 13 12",
            "card 3 0 11 10 9",
            "end",
            "skill 1 15 14 13",
            "TEST 1 10 4 10 0 1 6",
            "skill 0 12 11 10",
            "skill 0 9 8 7",
            "TEST 1 10 4 10 0 0 2",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "end",
            "choose 1",
            "card 1 0 15 14 13",
            "sw_char 2 12",
            "sw_char 1 11",
            "skill 1 10 9 8",
            "end",
            "end",
            "choose 2",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Gorou
        character:Dvalin
        character:Alhaitham
        Rending Vortex*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_eremite_2():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 0 15 14 13",
            "end",
            "sw_char 1 15",
            "skill 0 14 13 12",
            "skill 0 11 10 9",
            "skill 0 8 7 6",
            "sw_char 0 5",
            "end",
            "choose 1",
            "end",
            "TEST 2 p0c0 charge 0",
            "sw_char 2 15",
            "end",
            "end",
            "TEST 1 0 2 0 10 4 9",
            "choose 1",
        ],
        [
            "sw_card",
            "choose 2",
            "skill 0 15 14 13",
            "sw_char 1 12",
            "end",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "skill 0 9 8 7",
            "end",
            "skill 2 15 14 13",
            "sw_char 0 12",
            "skill 0 11 10 9",
            "skill 0 8 7 6",
            "end",
            "card 7 0 15 14 13",
            "sw_char 2 12",
            "skill 0 11 10 9",
            "TEST 3 p1 summon damage 1",
            "sw_char 0 8",
            "skill 0 7 6 5",
            "skill 0 4 3 2",
            "end",
            "skill 2 15 14 13",
            "TEST 3 p1 summon damage 1",
            "TEST 1 0 2 1 10 4 9",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Eremite Scorching Loremaster
        character:Chongyun
        character:Klee
        Scorpocalypse*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                assert match.player_tables[0].characters[0].charge == 0
            elif test_id == 3:
                assert match.player_tables[1].summons[0].damage == 1
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_signora_2():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 0 15 14 13",
            "skill 2 12 11 10",
            "TEST 2 p1c0 usage 1 1",
            "skill 4 9 8 7",
            "TEST 2 p1c0 usage 1 1",
            "skill 3 6 5 4",
            "TEST 2 p1c0 usage 1",
            "end",
        ],
        ["sw_card", "choose 0", "end"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Signora
        character:Chongyun
        character:Klee
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 2:
                status = match.player_tables[1].characters[0].status
                check_usage(status, cmd[4:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_thunder_2():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 1 0",
            "skill 1 15 14 13",
            "sw_char 1 12",
            "skill 0 11 10 9",
            "TEST 1 10 6 10 10 7 7",
            "TEST 2 p0 hand 5",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "TEST 1 10 10 10 10 7 10",
            "sw_char 2 15",
            "skill 0 14 13 12",
            "skill 1 11 10 9",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Thunder Manifestation
        character:Raiden Shogun
        character:Lyney
        Grieving Echo*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                assert len(match.player_tables[0].hands) == 5
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_jade_razor():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 1 0 15 14 13",
            "skill 0 12 11 10",
            "card 1 0",
            "card 1 0 9",
            "skill 0 8 7 6",
            "TEST 1 10 10 10 10 4 10",
            "skill 0 5 4 3",
        ],
        ["sw_card", "choose 1", "TEST 1 10 10 10 10 7 10", "end", "choose 0"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Shenhe
        character:Raiden Shogun
        character:Lyney
        Primordial Jade Winged-Spear*10
        Where Is the Unseen Razor?*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # test_lynette_single()
    test_gorou_alhaitham_layla()
    # test_dvalin_2()
    # test_eremite_2()
    # test_signora_2()
    # test_thunder_2()



import time
import json
from typing import Literal
import dictdiffer
import pytest

from lpsim.server.struct import Cost, ObjectPosition
from lpsim.server.event import UseCardEventArguments
from lpsim.server.action import ActionTypes, UseCardAction
from lpsim.utils.class_registry import get_instance
from lpsim.server.summon.base import SummonBase
from lpsim.server.consts import ObjectPositionType, ObjectType
from lpsim.agents.nothing_agent import NothingAgent
from lpsim.server.event_handler import OmnipotentGuideEventHandler_3_3
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from lpsim.agents.random_agent import RandomAgent
from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.server.character.electro.fischl_3_3 import Oz_3_3
from tests.utils_for_test import (
    get_test_id_from_command,
    set_16_omni,
    make_respond,
    remove_ids,
    get_random_state,
    check_hp,
)


TEST_DECK = {
    "name": "Deck",
    "characters": [
        {
            "name": "DendroMobMage",
            "element": "DENDRO",
            "hp": 99,
            "max_hp": 99,
        },
        {
            "name": "DendroMobMage",
            "element": "DENDRO",
        },
        {
            "name": "ElectroMobMage",
            "element": "ELECTRO",
        },
    ],
    "cards": [
        {
            "name": "Strategize",
        }
    ]
    * 30,
}


def test_match_pipeline():
    agent_0 = RandomAgent(player_idx=0)
    agent_1 = NothingAgent(player_idx=1)
    match = Match()
    deck = TEST_DECK
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    set_16_omni(match)
    # main.match.event_handlers[0].version = '3.3'
    assert match.start()[0]
    match.step()  # switch card

    while match.round_number < 100 and not match.is_game_end():
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)

    assert match.state != MatchState.ERROR


def test_save_load():
    """
    use v3.3 Catalyzing Field to test save & load.
    when save on field generated, load should be same. otherwise, load should
    be different only on id.
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            # 'reroll',
            "skill 0 omni omni omni",
            "sw_char 0 omni",
            "skill 0 omni omni omni",  # quiken activated
            "skill 0 omni omni omni",
            "sw_char 2 omni",
            "skill 0 omni omni omni",  # quiken activated, renew in v3.3
            "end",
            "skill 0 omni omni omni",
            "skill 0 omni omni omni",
            "skill 0 omni omni omni",  # run out of field
        ],
        only_use_command=True,
    )
    match = Match()
    deck = TEST_DECK
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    set_16_omni(match)
    match.event_handlers[0].version = "3.3"  # type: ignore
    assert match.start()[0]
    match.step()  # switch card

    saves = {}
    agent_saves = {}

    while True:
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)
        if len(agent_1.commands) not in saves:
            length = len(agent_1.commands)
            saves[length] = match.copy(deep=True)
            agent_saves[length] = agent_1.copy(deep=True)
        else:
            raise AssertionError()
        if len(agent_1.commands) == 0:
            break

    def check_range(saves, agent_saves, start, end, same=True, ignore_id=False):
        """
        load start and run until end, then check if they are same.
        if ignore_id, before compare will remove ids from objects.
        """
        match = saves[start].copy(deep=True)
        agent_1 = agent_saves[start].copy(deep=True)
        while True:
            make_respond(agent_0, match, assertion=False)
            make_respond(agent_1, match, assertion=False)
            if len(agent_1.commands) == end:
                break
        saves_end = saves[end]
        if ignore_id:
            match = match.copy(deep=True)
            saves_end = saves_end.copy(deep=True)
            remove_ids(match)
            remove_ids(saves_end)
        is_same = match == saves_end
        assert is_same == same

    # from any position, run reuslts should same
    for _ in range(1, 11):
        check_range(saves, agent_saves, _, 0, same=True, ignore_id=True)
    # generate status, id should not change
    check_range(saves, agent_saves, 8, 7, same=False, ignore_id=False)
    # ignore id, should same
    check_range(saves, agent_saves, 8, 7, same=True, ignore_id=True)
    # renew status, id should not change
    check_range(saves, agent_saves, 6, 4, same=True, ignore_id=False)

    assert len(agent_1.commands) == 0

    assert match.state != MatchState.ERROR


def test_copy_speed():
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "card 1 0 0 1",
            "skill 1 0 1 2",
            "sw_char 1 1",
            "skill 1 omni omni hydro",
            "end",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "skill 2 omni omni dendro omni omni",
            "end",
            "skill 1 0 1 2",
            "card 0 0 0 1 2",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:Fischl
        character:Mona
        character:Nahida
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert match.state != MatchState.ERROR

    run_time = 100
    match = Match()
    deck = TEST_DECK
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    assert match.start()[0]
    starttime = time.time()
    for i in range(run_time):
        _ = match.copy(deep=True)
    endtime = time.time()
    print("not json speed", (endtime - starttime) / run_time)

    match = Match()
    match.config.max_same_card_number = 30
    match.set_deck([deck, deck])
    mainjson = match.json()
    starttime = time.time()
    assert match.start()[0]
    for i in range(run_time):
        mainjson = match.json()
        _ = Match(**json.loads(mainjson))
    endtime = time.time()
    print("json speed", (endtime - starttime) / run_time)
    assert True


def test_random_same_after_load():
    agent_0 = RandomAgent(player_idx=0, random_seed=42)
    agent_1 = RandomAgent(player_idx=1, random_seed=19260817)
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Rhodeia of Loch
        character:Kamisato Ayaka
        Traveler's Handy Sword*5
        Gambler's Earrings*5
        Kanten Senmyou Blessing*5
        Sweet Madame*5
        Abyssal Summons*5
        Fatui Conspiracy*5
        """
    )
    for character in deck.characters:
        character.hp = character.max_hp = 99
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.max_hand_size = 30
    match.config.initial_hand_size = 10
    match.config.card_number = None
    match.config.character_number = None
    match.config.check_deck_restriction = False
    initial_match = match.copy(deep=True)
    initial_agent_0 = agent_0.copy(deep=True)
    initial_agent_1 = agent_1.copy(deep=True)
    results_1 = []
    test_step = 100
    assert match.start()[0]
    match.step()
    for _ in range(test_step):
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)
        results_1.append(match.copy(deep=True))
    assert match.state != MatchState.ERROR

    # use deepcopy to rerun
    match = initial_match.copy(deep=True)
    agent_0 = initial_agent_0.copy(deep=True)
    agent_1 = initial_agent_1.copy(deep=True)
    assert match.start()[0]
    match.step()
    for i in range(test_step):
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)
        remove_ids(results_1[i])
        now = match.copy(deep=True)
        remove_ids(now)
        assert now == results_1[i], f"deepcopy error at {i}"
    assert match.state != MatchState.ERROR

    match = Match(**json.loads(initial_match.json()))
    assert match.json() == initial_match.json()
    assert match == initial_match
    # random agent do not support json, use copy instead
    agent_0 = initial_agent_0.copy()
    agent_1 = initial_agent_1.copy()
    assert agent_0 == initial_agent_0
    assert agent_1 == initial_agent_1
    assert match.start()[0]
    match.step()
    for i in range(test_step):
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)
        remove_ids(results_1[i])
        now = match.copy(deep=True)
        remove_ids(now)
        assert now == results_1[i], f"json error at {i}"
    assert match.state != MatchState.ERROR


def test_save_load_same():
    agent_0 = RandomAgent(player_idx=0, random_seed=42)
    agent_1 = RandomAgent(player_idx=1, random_seed=19260817)
    match = Match(random_state=get_random_state(100))
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Rhodeia of Loch
        character:Kamisato Ayaka
        Traveler's Handy Sword*5
        Gambler's Earrings*5
        Kanten Senmyou Blessing*5
        Sweet Madame*5
        Abyssal Summons*5
        Fatui Conspiracy*5
        Timmie*5
        """
    )
    for character in deck.characters:
        character.hp = character.max_hp = 99
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.max_hand_size = 30
    match.config.initial_hand_size = 10
    match.config.card_number = None
    match.config.character_number = None
    match.config.check_deck_restriction = False
    test_step = 100
    assert match.start()[0]
    match.step()
    for _ in range(test_step):
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)
        match_copy = match.copy(deep=True)
        assert remove_ids(match_copy) == remove_ids(
            Match(**json.loads(match_copy.json()))
        )
        match = Match(**json.loads(match.json()))
    assert match.state != MatchState.ERROR


def test_use_card():
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        commands=[
            "sw_card",
            "choose 2",
            "card 0 0 omni",
            "card 0 0 omni",
            "card 0 0 omni",
            "card 0 0 omni",
            "end",
            "card 0 0 omni",
            "card 0 0 omni",
            "card 0 0 omni",
            "card 0 0 omni",
        ],
    )
    match = Match()
    deck = TEST_DECK
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.initial_dice_number = 16
    match.event_handlers = [OmnipotentGuideEventHandler_3_3()] + match.event_handlers
    assert match.start()[0]
    match.step()

    hand_numbers = [10, 10, 10, 10, 10, 9, 8, 7, 6, 5, 5, 5]
    deck_numbers = [7, 9, 11, 13, 15, 17, 19, 21, 23, 25, 25, 25]

    while len(agent_1.commands) > 0:
        if match.need_respond(0):
            current_agent = agent_0
        elif match.need_respond(1):
            current_agent = agent_1
            assert (
                len(match.player_tables[1].hands) == hand_numbers[len(agent_1.commands)]
            )
            assert (
                len(match.player_tables[1].table_deck)
                == deck_numbers[len(agent_1.commands)]
            )
        else:
            raise AssertionError("no need respond")
        make_respond(current_agent, match)

    assert match.state != MatchState.ERROR


def test_support_over_maximum_and_error_tests():
    """
    put 6 Ranas and check usages and order.
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 1",
            "card 0 0 omni omni",
            "card 0 0 omni omni",
            "card 0 0 omni omni",
            "end",
            "end",
            "skill 1 omni omni omni",
            "card 0 0 omni electro",
            "card 0 3 electro electro",
            # 4 rana, usage 0 0 0 1
            "card 0 0 omni omni",
            # 4 rana, usage 0 0 1 1
            "skill 1 omni omni omni",
            # dice 7 omni + 2 electro
            "end",
            "skill 1 omni omni omni",
            # 13 omni and 3 electro and 0 usage
            "end",
            "skill 2 omni omni omni",
            # 13 omni, all 1 usage
            "end",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = {
        "name": "Deck",
        "characters": [
            {
                "name": "DendroMobMage",
                "element": "DENDRO",
                "hp": 99,
                "max_hp": 99,
            },
            {
                "name": "DendroMobMage",
                "element": "DENDRO",
            },
            {
                "name": "ElectroMobMage",
                "element": "ELECTRO",
            },
        ],
        "cards": [
            {
                "name": "Rana",
            }
        ]
        * 30,
    }
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            # asserts
            if len(agent_1.commands) == 7:
                # 4 rana, usage 0 0 0 1
                assert len(match.player_tables[1].supports) == 4
                for support, usage in zip(
                    match.player_tables[1].supports, [0, 0, 0, 1]
                ):
                    assert support.name == "Rana"
                    assert support.usage == usage
            elif len(agent_1.commands) == 6:
                # 4 rana, usage 0 0 1 1
                assert len(match.player_tables[1].supports) == 4
                for support, usage in zip(
                    match.player_tables[1].supports, [0, 0, 1, 1]
                ):
                    assert support.name == "Rana"
                    assert support.usage == usage
            elif len(agent_1.commands) == 5:
                # dice 7 omni + 2 electro
                assert len(match.player_tables[1].dice.colors) == 9
                omni_counter = 0
                electro_counter = 0
                for d in match.player_tables[1].dice.colors:
                    if d == "OMNI":
                        omni_counter += 1
                    else:
                        assert d == "ELECTRO"
                        electro_counter += 1
                assert omni_counter == 7
                assert electro_counter == 2
            elif len(agent_1.commands) == 3:
                # 13 omni and 3 electro and 0 usage
                assert len(match.player_tables[1].dice.colors) == 16
                omni_counter = 0
                electro_counter = 0
                for d in match.player_tables[1].dice.colors:
                    if d == "OMNI":
                        omni_counter += 1
                    else:
                        assert d == "ELECTRO"
                        electro_counter += 1
                assert omni_counter == 13
                assert electro_counter == 3
                for support in match.player_tables[1].supports:
                    assert support.name == "Rana"
                    assert support.usage == 0
            elif len(agent_1.commands) == 1:
                # 13 omni, all 1 usage
                assert len(match.player_tables[1].dice.colors) == 13
                for d in match.player_tables[1].dice.colors:
                    assert d == "OMNI"
                for support in match.player_tables[1].supports:
                    assert support.name == "Rana"
                    assert support.usage == 1

            make_respond(agent_1, match)
        else:
            raise AssertionError("no need respond")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 6
    check_hp(match, [[85, 10, 10], [99, 10, 10]])
    assert len(match.player_tables[1].supports) == 4
    for support in match.player_tables[1].supports:
        assert support.name == "Rana"

    assert match.state != MatchState.ERROR

    # after start, set deck will raise error
    with pytest.raises(ValueError):
        match.set_deck([deck, deck])
    assert not match.start()[0]


def test_summon_over_maximum():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "skill 2 0 1 2 3 4",
            "skill 2 0 1 2 3 4",
            "sw_char 1 0",
            "skill 1 0 1 2",
            "end",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "end",
        ],
        [
            "sw_card",
            "choose 2",
            "skill 2 0 1 2 3 4",
            "skill 2 0 1 2 3 4",
            "sw_char 1 0",
            "skill 1 0 1 2",
            "end",
            "sw_char 0 0",
            "sw_char 2 0",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "skill 2 0 1 2 3 4",
            "TEST 1 frog squi rap oz rap squi oz ref",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Mona
        character:Fischl
        character:Rhodeia of Loch
        Send Off*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.split()
                summons = cmd[-8:]
                summons = [summons[:4], summons[4:]]
                for table, name in zip(match.player_tables, summons):
                    assert len(table.summons) == 4
                    for s, n in zip(table.summons, name):
                        assert n in s.name.lower()
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_plunge_mark():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "TEST 1 p0 true",
            "TEST 1 p1 true",
            "skill 1 0 1 2",
            "TEST 1 p1 true",
            "sw_char 2 0",
            "TEST 1 p1 false",
            "card 0 0 0 1",
            "TEST 1 p0 true",
            "skill 0 0 1 2",
        ],
        [
            "sw_card",
            "choose 0",
            "TEST 1 p0 false",
            "sw_char 1 0",
            "TEST 1 p0 true",
            "skill 1 0 1 2",
            "TEST 1 p0 false",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida
        character:Fischl
        character:Rhodeia of Loch
        Send Off*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                cmd = cmd.split()
                b = bool(["false", "true"].index(cmd[-1].lower()))
                pid = int(cmd[2][1])
                assert match.player_tables[pid].plunge_satisfied == b
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_higher_version_compatible():
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Yoimiya@3.3
        character:Yoimiya@3.4
        character:Yoimiya@3.5
        character:Yoimiya@3.6
        character:Yoimiya@3.7
        character:Yoimiya@3.8
        character:Yoimiya@4.0
        character:Yoimiya
        Send Off@3.3
        Send Off@3.4
        Send Off@3.5*1
        Send Off@3.6*1
        Send Off@3.7*1
        Send Off@3.8*1
        Send Off@4.0*1
        Send Off*1
        """
    )
    res = ["3.3", "3.4", "3.4", "3.4", "3.4", "3.8", "3.8", "3.8"]
    res2 = ["3.3", "3.3", "3.3", "3.3", "3.7", "3.7", "3.7", "3.7"]
    for c, r in zip(deck.characters, res):
        assert c.version == r
    for c, r in zip(deck.cards, res2):
        assert c.version == r


def test_save_history():
    agent_0 = RandomAgent(player_idx=0, random_seed=42)
    agent_1 = RandomAgent(player_idx=1, random_seed=19260817)
    match = Match(random_state=get_random_state(100))
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Rhodeia of Loch
        character:Kamisato Ayaka
        Traveler's Handy Sword*5
        Gambler's Earrings*5
        Kanten Senmyou Blessing*5
        Sweet Madame*5
        Abyssal Summons*5
        Fatui Conspiracy*5
        Timmie*5
        """
    )
    for character in deck.characters:
        character.hp = character.max_hp = 99
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.max_hand_size = 30
    match.config.initial_hand_size = 10
    match.config.card_number = None
    match.config.character_number = None
    match.config.check_deck_restriction = False
    match.config.history_level = 10  # record important history
    match.config.compress_history = False
    test_step = 100
    assert match.start()[0]
    match.step()
    for _ in range(test_step):
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)
    assert len(match._history) > test_step  # should record history
    match_dict = match._history[0].copy(deep=True).dict()
    starttime = time.time()
    for patch in match._history_diff[1:]:
        match_dict = dictdiffer.patch(patch, match_dict)
    print("patch speed", (time.time() - starttime) / len(match._history_diff))
    assert match._history[-1].dict() == match_dict

    match_dict = match._history[0].copy(deep=True).dict()
    starttime = time.time()
    for patch in match._history_diff[1:]:
        dictdiffer.patch(patch, match_dict, in_place=True)
    print("patch speed in-place", (time.time() - starttime) / len(match._history_diff))
    assert match._history[-1].dict() == match_dict

    # for prev, next, diff in zip(match._history[:-1], match._history[1:],
    #                             match._history_diff[1:]):
    #     assert next.dict() == dictdiffer.patch(diff, prev.dict())
    #     assert next.dict() == Match(**next.dict()).dict()
    #     assert next == Match(**dictdiffer.patch(diff, prev.dict()))
    assert match.state != MatchState.ERROR


def test_save_history_2():
    agent_0 = RandomAgent(player_idx=0, random_seed=42)
    agent_1 = RandomAgent(player_idx=1, random_seed=19260817)
    match = Match(random_state=get_random_state(100))
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Rhodeia of Loch
        character:Kamisato Ayaka
        Toss-Up*10
        Nature and Wisdom*10
        """
    )
    for character in deck.characters:
        character.hp = character.max_hp = 1
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.max_hand_size = 30
    match.config.initial_hand_size = 10
    match.config.card_number = None
    match.config.character_number = None
    match.config.check_deck_restriction = False
    match.config.history_level = 10  # record important history
    match.config.compress_history = False
    test_step = 10
    assert match.start()[0]
    match.step()
    for _ in range(test_step):
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)
    assert len(match._history) > test_step  # should record history
    assert match.state != MatchState.ERROR


def test_generate_unused_cards():
    agent_0 = RandomAgent(player_idx=0, random_seed=42)
    agent_1 = RandomAgent(player_idx=1, random_seed=19260817)
    match = Match(random_state=get_random_state(100))
    deck = Deck.from_str(
        """
        default_version:4.0
        character:AnemoMob
        character:PyroMob
        character:DendroMob
        character:ElectroMob
        character:HydroMob
        character:CryoMob
        character:GeoMob
        Elemental Resonance: Woven Stone
        Skyward Harp
        """
    )
    for character in deck.characters:
        character.hp = character.max_hp = 1
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.max_hand_size = 30
    match.config.initial_hand_size = 10
    match.config.card_number = None
    match.config.character_number = None
    match.config.check_deck_restriction = False
    match.config.history_level = 10  # record important history
    test_step = 10
    assert match.start()[0]
    match.step()
    for _ in range(test_step):
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)
    assert len(match._history_diff) > test_step  # should record history
    assert match.state != MatchState.ERROR


def test_prediction():
    """
    3335 + E + talent
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "card 1 0 0 1",
            "skill 1 0 1 2",
            "sw_char 1 1",
            "skill 1 omni omni hydro",
            "end",
            "TEST 1 hp 4 8 8",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "skill 2 omni omni dendro omni omni",
            "TEST 2 hp 0 7 7 and 1 seed for alive",
            "end",
            "TEST 3 hp 0 1 5",
            "skill 1 0 1 2",
            "card 0 0 0 1 2",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:Fischl
        character:Mona
        character:Nahida
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.make_skill_prediction = True
    match.config.history_level = 10
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    check_hp(match, [[4, 8, 8], [10, 10, 10]])
                elif test_id == 2:
                    check_hp(match, [[0, 7, 7], [10, 10, 10]])
                    for character in match.player_tables[0].characters:
                        if character.is_alive:
                            assert len(character.status) == 1
                            assert character.status[0].name == "Seed of Skandha"
                        else:
                            assert len(character.status) == 0
                elif test_id == 3:
                    check_hp(match, [[0, 1, 5], [10, 10, 10]])
                else:
                    break
            prediction_found = None
            prediction_all = None
            if len(agent_1.commands) > 0 and agent_1.commands[0][:5] == "skill":
                # is skill, check prediction
                prediction_all = match.skill_predictions
                for p in prediction_all:
                    assert p["player_idx"] == 1
                    assert (
                        p["character_idx"]
                        == match.player_tables[1].active_character_idx
                    )
                    skill_idx = int(agent_1.commands[0].split()[1])
                    if skill_idx == p["skill_idx"]:
                        prediction_found = p
            if prediction_all is not None:
                assert prediction_found is not None
                prediction_found["current_dict"] = match.dict(
                    exclude={"skill_predictions"}
                )
            make_respond(agent_1, match)
            after_dict = match.dict(exclude={"skill_predictions"})
            if prediction_found is not None:
                patch_dict = dictdiffer.patch(
                    prediction_found["diff"], prediction_found["current_dict"]
                )
                # remove ids, as newly created ids may differ
                patch_match = Match(**patch_dict)
                after_match = Match(**after_dict)
                remove_ids(patch_match)
                remove_ids(after_match)
                for d in (patch_match, after_match):
                    for tb in d.player_tables:
                        # remove dice, as prediction will not use dice
                        tb.dice.colors = []
                        tb.dice.colors = []
                        # as dice number may differ, charge may differ
                        tb.charge_satisfied = False
                    # remove requests, as prediction will not use requests
                    d.requests = []
                    # remove event frames, as it contains ids in some action,
                    # which is hard to remove
                    d.event_frames = []
                assert after_match == patch_match
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[0, 0, 1], [10, 10, 10]])
    assert len(match.player_tables[1].team_status) == 1
    assert match.player_tables[1].team_status[0].name == "Shrine of Maya"
    assert match.player_tables[1].team_status[0].usage == 3

    assert match.state != MatchState.ERROR


def test_player_table_character_order():
    deck_str = """
    default_version:4.0
    character:Fischl
    character:Mona
    character:Nahida
    Strategize*30
    """
    deck = Deck.from_str(deck_str)
    match = Match(random_state=get_random_state())
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    assert match.start()[0]
    table = match.player_tables[0].copy(deep=True)
    for active_idx in range(3):
        table.active_character_idx = active_idx
        objs = table.get_object_lists()
        result = []
        for obj in objs:
            if obj.type == ObjectType.CHARACTER:
                result.append(obj.position.character_idx)
        if active_idx == 0:
            assert result == [0, 1, 2]
        elif active_idx == 1:
            assert result == [1, 2, 0]
        elif active_idx == 2:
            assert result == [2, 0, 1]
        else:
            raise AssertionError()


def test_frozen_cannot_use_skill_talent():
    cmd_records = [
        ["sw_card", "choose 2", "skill 1 15 14 13", "sw_char 1 12", "skill 1 11 10 9"],
        ["sw_card", "choose 1", "skill 1 15 14 13", "TEST 1 no card can use", "end"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Electro Hypostasis
        character:Kaeya
        character:Mona
        Cold-Blooded Strike*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                for req in match.requests:
                    assert req.name != "UseCardRequest"
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_new_match_from_history():
    agent_0 = RandomAgent(player_idx=0, random_seed=42)
    agent_1 = RandomAgent(player_idx=1, random_seed=19260817)
    match = Match(random_state=get_random_state(100))
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Rhodeia of Loch
        character:Kamisato Ayaka
        Traveler's Handy Sword*5
        Gambler's Earrings*5
        Kanten Senmyou Blessing*5
        Sweet Madame*5
        Abyssal Summons*5
        Fatui Conspiracy*5
        Timmie*5
        """
    )
    for character in deck.characters:
        character.hp = character.max_hp = 99
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.max_hand_size = 30
    match.config.initial_hand_size = 10
    match.config.card_number = None
    match.config.character_number = None
    match.config.check_deck_restriction = False
    match.config.history_level = 10  # record important history
    match.config.compress_history = False
    assert match.start()[0]
    match.step()
    for _ in range(10):
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)
    idx_10 = len(match._history) - 1
    for _ in range(10):
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)
    assert len(match._history) > 20  # should record history
    initial_match = match.new_match_from_history(0)
    match_10 = match.new_match_from_history(idx_10)
    agent_0 = RandomAgent(player_idx=0, random_seed=42)
    agent_1 = RandomAgent(player_idx=1, random_seed=19260817)
    initial_match.step()
    for _ in range(10):
        make_respond(agent_0, initial_match, assertion=False)
        make_respond(agent_1, initial_match, assertion=False)
    assert len(initial_match._history) == len(match_10._history)
    assert len(initial_match._history_diff) == len(match_10._history_diff)
    assert remove_ids(initial_match.copy(deep=True)) == remove_ids(match_10)
    for _ in range(10):
        make_respond(agent_0, initial_match, assertion=False)
        make_respond(agent_1, initial_match, assertion=False)
    assert len(initial_match._history) == len(match._history)
    assert len(initial_match._history_diff) == len(match._history_diff)
    assert remove_ids(initial_match) == remove_ids(match)


def test_new_match_from_history_compressed():
    agent_0 = RandomAgent(player_idx=0, random_seed=42)
    agent_1 = RandomAgent(player_idx=1, random_seed=19260817)
    match = Match(random_state=get_random_state(100))
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Rhodeia of Loch
        character:Kamisato Ayaka
        Traveler's Handy Sword*5
        Gambler's Earrings*5
        Kanten Senmyou Blessing*5
        Sweet Madame*5
        Abyssal Summons*5
        Fatui Conspiracy*5
        Timmie*5
        """
    )
    for character in deck.characters:
        character.hp = character.max_hp = 99
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.max_hand_size = 30
    match.config.initial_hand_size = 10
    match.config.card_number = None
    match.config.character_number = None
    match.config.check_deck_restriction = False
    match.config.history_level = 10  # record important history
    assert match.start()[0]
    match.step()
    for _ in range(10):
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)
    idx_10 = len(match._history_diff) - 1
    for _ in range(10):
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)
    assert len(match._history_diff) > 20  # should record history
    initial_match = match.new_match_from_history(0)
    match_10 = match.new_match_from_history(idx_10)
    agent_0 = RandomAgent(player_idx=0, random_seed=42)
    agent_1 = RandomAgent(player_idx=1, random_seed=19260817)
    initial_match.step()
    for _ in range(10):
        make_respond(agent_0, initial_match, assertion=False)
        make_respond(agent_1, initial_match, assertion=False)
    assert len(initial_match._history) == len(match_10._history)
    assert len(initial_match._history_diff) == len(match_10._history_diff)
    assert remove_ids(initial_match.copy(deep=True)) == remove_ids(match_10)
    for _ in range(10):
        make_respond(agent_0, initial_match, assertion=False)
        make_respond(agent_1, initial_match, assertion=False)
    assert remove_ids(initial_match) == remove_ids(match)
    assert len(initial_match._history) == len(match._history)
    assert len(initial_match._history_diff) == len(match._history_diff)


def test_version_validation():
    with pytest.raises(ValueError):
        _ = get_instance(SummonBase, {"name": "Oz", "version": 3.7})
    with pytest.raises(AssertionError):
        _ = get_instance(SummonBase, {"name": "Oz", "version": "2.5"})
    with pytest.raises(ValueError):
        _ = get_instance(
            SummonBase,
            {"name": "Oz", "strict_version_validation": True, "version": "3.5"},
        )
    default_pos = {"player_idx": 0, "id": -1, "area": "SUMMON"}
    _ = get_instance(
        SummonBase,
        {
            "name": "Oz",
            "version": "3.5",
            "position": default_pos,
        },
    )
    with pytest.raises(ValueError):
        _ = Oz_3_3(
            **{
                "name": "Oz",
                "version": 3.7,
            }
        )
    with pytest.raises(ValueError):
        _ = Oz_3_3(
            **{
                "name": "Oz",
                "version": "2.5",
                "position": default_pos,
            }
        )
    with pytest.raises(ValueError):
        _ = Oz_3_3(
            **{
                "name": "Oz",
                "strict_version_validation": True,
                "version": "3.5",
                "position": default_pos,
            }
        )
    _ = Oz_3_3(
        **{
            "name": "Oz",
            "version": "3.5",
            "position": default_pos,
        }
    )

    # multiple version hints
    class OOZZ(Oz_3_3):
        version: Literal["3.4", "3.5"]

    with pytest.raises(ValueError):
        _ = OOZZ(
            **{
                "name": "Oz",
                "version": "3.5",
                "position": default_pos,
            }
        )


def test_round_end_all_lose():
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = NothingAgent(player_idx=1)
    match = Match()
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Rhodeia of Loch
        character:Kamisato Ayaka
        Traveler's Handy Sword*5
        Gambler's Earrings*5
        Kanten Senmyou Blessing*5
        Sweet Madame*5
        Abyssal Summons*5
        Fatui Conspiracy*5
        Timmie*5
        """
    )
    round_number = 5
    match.set_deck([deck, deck])
    match.config.max_round_number = round_number
    match.config.max_same_card_number = 30
    match.config.max_hand_size = 30
    match.config.initial_hand_size = 10
    match.config.card_number = None
    match.config.character_number = None
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    match.config.history_level = 0
    set_16_omni(match)
    assert match.start()[0]
    match.step()
    for i in range(round_number + 1):  # +1 to perform sw_card and choose
        assert match.state not in [MatchState.ENDED, MatchState.ERROR]
        make_respond(agent_0, match, assertion=False)
        make_respond(agent_1, match, assertion=False)
        target_round_number = max(i, 0)
        assert match.round_number == target_round_number
    assert match.state == MatchState.ENDED


def test_use_card_event_serialize():
    card_event = UseCardEventArguments(
        action=UseCardAction(
            card_position=ObjectPosition(
                player_idx=0, area=ObjectPositionType.HAND, id=0
            ),
            target=None,
        ),
        card_cost=Cost(),
        card_name="Test",
        use_card_success=True,
    )
    assert card_event.card_cost.original_value is None
    card_event = UseCardEventArguments(**json.loads(card_event.json()))
    assert card_event.card_cost.original_value is None
    assert card_event.action.type == ActionTypes.USE_CARD
    assert card_event.action.card_position.area == ObjectPositionType.HAND


if __name__ == "__main__":
    # test_match_pipeline()
    # test_save_load()
    # test_random_same_after_load()
    # test_use_card()
    # test_copy_speed()
    # test_save_load_same()
    # test_support_over_maximum_and_error_tests()
    # test_summon_over_maximum()
    # test_plunge_mark()
    # test_higher_version_compatible()
    # test_save_history()
    # test_save_history_2()
    # test_generate_unused_cards()
    # test_prediction()
    # test_player_table_character_order()
    # test_frozen_cannot_use_skill_talent()
    # test_new_match_from_history()
    # test_new_match_from_history_compressed()
    # test_version_validation()
    # test_round_end_all_lose()
    test_use_card_event_serialize()



from typing import List, Literal

from .struct import ObjectPosition
from .consts import DieColor, ObjectPositionType
from .object_base import ObjectBase, ObjectType


class Dice(ObjectBase):
    """
    Class representing dice.

    Attributes:
        colors: list of colors of dice.
    """

    name: Literal["Dice"] = "Dice"
    position: ObjectPosition = ObjectPosition(
        player_idx=-1,
        area=ObjectPositionType.INVALID,
        id=-1,
    )
    type: Literal[ObjectType.DICE] = ObjectType.DICE
    colors: List[DieColor] = []

    def colors_to_idx(self, colors: List[DieColor]) -> List[int]:
        """
        Convert colors to idx.
        """
        res: List[int] = []
        all_c: List[DieColor | None] = list(self.colors)
        for x in colors:
            res.append(all_c.index(x))
            all_c[all_c.index(x)] = None
        return res



from lpsim.server.consts import DieColor, ObjectPositionType
from lpsim.server.deck import Deck
from lpsim.server.interaction import (
    SwitchCardRequest,
    SwitchCardResponse,
    RerollDiceRequest,
    RerollDiceResponse,
    SwitchCharacterRequest,
    SwitchCharacterResponse,
    UseCardRequest,
    UseCardResponse,
    UseSkillRequest,
    UseSkillResponse,
)
from lpsim.server.match import Match
from lpsim.server.struct import Cost, ObjectPosition


def test_response_is_valid():
    # mini fake match
    match = Match()
    deck = Deck.from_str("character:Nahida")
    match.player_tables[0].characters.append(deck.characters[0])
    match.player_tables[1].characters.append(deck.characters[0])
    match.player_tables[0].active_character_idx = 0
    match.player_tables[1].active_character_idx = 0
    req = SwitchCharacterRequest(
        player_idx=0,
        active_character_idx=0,
        target_character_idx=1,
        dice_colors=[DieColor.CRYO, DieColor.PYRO],
        cost=Cost(
            any_dice_number=1,
        ),
    )
    resp = SwitchCharacterResponse(
        request=req,
        dice_idxs=[0],
    )
    assert resp.is_valid(match)
    resp.dice_idxs = []
    assert not resp.is_valid(match)
    resp.dice_idxs = [-1]
    assert not resp.is_valid(match)
    resp.dice_idxs = [0, 1]
    assert not resp.is_valid(match)
    resp.dice_idxs = [5]
    assert not resp.is_valid(match)
    resp.dice_idxs = []
    resp.request.cost.any_dice_number = 0
    assert resp.is_valid(match)
    req = RerollDiceRequest(
        player_idx=0,
        colors=[DieColor.CRYO, DieColor.PYRO, DieColor.PYRO, DieColor.PYRO],
        reroll_times=1,
    )
    resp = RerollDiceResponse(
        request=req,
        reroll_dice_idxs=[0, 1],
    )
    assert resp.is_valid(match)
    resp.reroll_dice_idxs = [0, 0]
    assert not resp.is_valid(match)
    resp.reroll_dice_idxs = [-1, 0]
    assert not resp.is_valid(match)
    resp.reroll_dice_idxs = [0, 4]
    assert not resp.is_valid(match)
    req = SwitchCardRequest(
        player_idx=0,
        card_names=["A", "B", "A", "B", "C"],
        maximum_switch_number=2,
    )
    resp = SwitchCardResponse(
        request=req,
        card_idxs=[0, 1],
    )
    assert resp.is_valid(match)
    resp.card_idxs = [0, 0]
    assert not resp.is_valid(match)
    resp.card_idxs = [-1, 0]
    assert not resp.is_valid(match)
    resp.card_idxs = [0, 5]
    assert not resp.is_valid(match)
    resp.card_idxs = [0, 1, 2, 3]
    assert not resp.is_valid(match)
    req = UseCardRequest(
        player_idx=0,
        card_idx=0,
        dice_colors=[DieColor.CRYO, DieColor.PYRO, DieColor.PYRO, DieColor.PYRO],
        targets=[],
        cost=Cost(
            same_dice_number=2,
        ),
    )
    resp = UseCardResponse(request=req, dice_idxs=[1, 2], target=None)
    assert resp.is_valid(match)
    resp.dice_idxs = [1, 1]
    assert not resp.is_valid(match)
    resp.dice_idxs = [-1, 1]
    assert not resp.is_valid(match)
    resp.dice_idxs = [1, 5]
    assert not resp.is_valid(match)
    resp.dice_idxs = [1, 2, 3]
    assert not resp.is_valid(match)
    resp.dice_idxs = [1, 2]
    target = ObjectPosition(
        player_idx=0,
        area=ObjectPositionType.HAND,
        id=-1,
    )
    req.targets.append(target)
    assert not resp.is_valid(match)
    req.targets.clear()
    resp.target = target
    assert not resp.is_valid(match)
    req.targets.append(target)
    assert resp.is_valid(match)
    resp.target = resp.target.set_id(321)
    assert not resp.is_valid(match)
    req = UseSkillRequest(
        player_idx=0,
        character_idx=2,
        skill_idx=1,
        dice_colors=[DieColor.CRYO, DieColor.PYRO, DieColor.PYRO, DieColor.PYRO],
        cost=Cost(
            elemental_dice_number=1,
            elemental_dice_color=DieColor.CRYO,
            any_dice_number=1,
        ),
    )
    resp = UseSkillResponse(
        request=req,
        dice_idxs=[0, 2],
    )
    assert resp.is_valid(match)
    resp.dice_idxs = []
    assert not resp.is_valid(match)
    resp.dice_idxs = [-1, 3]
    assert not resp.is_valid(match)
    resp.dice_idxs = [0, 0]
    assert not resp.is_valid(match)
    resp.dice_idxs = [0, 5]
    assert not resp.is_valid(match)
    resp.dice_idxs = [0, 1, 2]
    assert not resp.is_valid(match)
    resp.dice_idxs = [1, 2]
    assert not resp.is_valid(match)


if __name__ == "__main__":
    test_response_is_valid()



from typing import Dict, List, Literal

from .....utils.class_registry import register_class

from .....utils.desc_registry import DescDictType
from ....action import ConsumeArcaneLegendAction, DrawCardAction
from ....card.event.arcane_legend import ArcaneLegendBase
from ....match import Match
from ....struct import Cost, ObjectPosition


class InEveryHouseAStove_4_4(ArcaneLegendBase):
    name: Literal["In Every House a Stove"]
    version: Literal["4.4"] = "4.4"
    cost: Cost = Cost(arcane_legend=True)

    def get_targets(self, match: Match) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Match
    ) -> List[ConsumeArcaneLegendAction | DrawCardAction]:
        """
        Draw cards.
        """
        assert target is None
        return super().get_actions(target, match) + [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=min(match.round_number - 1, 4),
                draw_if_filtered_not_enough=True,
            )
        ]


desc: Dict[str, DescDictType] = {
    "ARCANE/In Every House a Stove": {
        "descs": {
            "4.4": {
                "zh-CN": "我方抓相当于当前的回合数减1的牌。（最多抓4张）（整局游戏只能打出一张「秘传」卡牌；这张牌一定在你的起始手牌中）",  # noqa: E501
                "en-US": 'Draw a number of cards equal to the current Round number minus 1. (Up to 4 cards can be drawn in this way) (Only one "Arcane Legend" card can be played for the entire game. This card will be in your starting hand.)',  # noqa: E501
            }
        },
    },
}


register_class(InEveryHouseAStove_4_4, desc)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions, RemoveObjectAction
from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)


# Skills


class InfluxBlast(ElementalSkillBase):
    name: Literal["Influx Blast"] = "Influx Blast"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        First remove existing Refraction on opposite, then create new
        Refraction.
        """
        target_characters = match.player_tables[1 - self.position.player_idx].characters
        exist_status = None
        for character in target_characters:
            for status in character.status:
                if status.name == "Refraction":
                    exist_status = status
                    break
            if exist_status is not None:
                break
        ret: List[RemoveObjectAction] = []
        if exist_status:
            ret.append(RemoveObjectAction(object_position=exist_status.position))
        args = {}
        if self.is_talent_equipped(match):
            args = {
                "usage": 3,
                "max_usage": 3,
                "is_talent_activated": True,
            }
        o_status = self.create_opposite_character_status(match, "Refraction", args)
        if len(ret) > 0 and o_status.object_position.check_position_valid(
            ret[0].object_position, match, player_idx_same=True, character_idx_same=True
        ):
            # attach to same character, do not remove
            ret = []
        return ret + super().get_actions(
            match, [self.create_opposite_character_status(match, "Refraction", args)]
        )


# Talents


class MirrorCage_3_3(SkillTalent):
    name: Literal["Mirror Cage"]
    version: Literal["3.3"]
    character_name: Literal["Mirror Maiden"] = "Mirror Maiden"
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=4)
    skill: Literal["Influx Blast"] = "Influx Blast"


class MirrorCage_4_2(MirrorCage_3_3):
    version: Literal["4.2"] = "4.2"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO,
        elemental_dice_number=3,
    )


# character base


class MirrorMaiden_3_7(CharacterBase):
    name: Literal["Mirror Maiden"]
    version: Literal["3.7"] = "3.7"
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[ElementalNormalAttackBase | InfluxBlast | ElementalBurstBase] = []
    faction: List[FactionType] = [FactionType.FATUI]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Water Ball",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            InfluxBlast(),
            ElementalBurstBase(
                name="Rippled Reflection",
                damage=5,
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalBurstBase.get_cost(self.element, 3, 2),
            ),
        ]


register_class(MirrorMaiden_3_7 | MirrorCage_4_2 | MirrorCage_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import SwirlChangeSummonBase

from ...modifiable_values import DamageIncreaseValue

from ...action import ActionTypes, Actions
from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class LargeWindSpirit_3_3(SwirlChangeSummonBase):
    name: Literal["Large Wind Spirit"] = "Large Wind Spirit"
    desc: Literal["", "talent"] = ""
    version: Literal["3.3"] = "3.3"
    usage: int = 3
    max_usage: int = 3
    damage: int = 2
    talent_activated: bool = False

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self.talent_activated:
            self.desc = "talent"

    def renew(self, obj: "LargeWindSpirit_3_3") -> None:
        super().renew(obj)
        if obj.talent_activated and not self.talent_activated:
            self.desc = "talent"
            self.talent_activated = obj.talent_activated

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        """
        If talent activated, and damage type is not anemo, increase damage
        """
        if (
            not self.talent_activated
            or self.damage_elemental_type == DamageElementalType.ANEMO
        ):
            # no talent or current is anemo
            return value
        if (
            value.position.player_idx != self.position.player_idx
            or value.target_position.player_idx == self.position.player_idx
        ):
            # enemy or attack self
            return value
        if value.damage_elemental_type != self.damage_elemental_type:
            # elemental type not match
            return value
        # increase damage
        assert mode == "REAL"
        value.damage += 1
        return value


# Skills


class AstableAnemohypostasisCreation6308(ElementalSkillBase):
    name: Literal[
        "Astable Anemohypostasis Creation - 6308"
    ] = "Astable Anemohypostasis Creation - 6308"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        prev_idx = match.player_tables[
            1 - self.position.player_idx
        ].previous_character_idx()
        if prev_idx is None:
            return ret
        # change character
        attack_action = ret[0]
        assert attack_action.type == ActionTypes.MAKE_DAMAGE
        attack_action.character_change_idx[1 - self.position.player_idx] = prev_idx
        return ret


class ForbiddenCreationIsomer75TypeII(ElementalBurstBase):
    name: Literal[
        "Forbidden Creation - Isomer 75 / Type II"
    ] = "Forbidden Creation - Isomer 75 / Type II"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        args = {"talent_activated": self.is_talent_equipped(match)}
        return super().get_actions(
            match, [self.create_summon("Large Wind Spirit", args)]
        )


# Talents


class ChaoticEntropy_3_3(SkillTalent):
    name: Literal["Chaotic Entropy"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Sucrose"] = "Sucrose"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3, charge=2
    )
    skill: Literal[
        "Forbidden Creation - Isomer 75 / Type II"
    ] = "Forbidden Creation - Isomer 75 / Type II"


# character base


class Sucrose_3_3(CharacterBase):
    name: Literal["Sucrose"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase
        | AstableAnemohypostasisCreation6308
        | ForbiddenCreationIsomer75TypeII
    ] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Wind Spirit Creation",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            AstableAnemohypostasisCreation6308(),
            ForbiddenCreationIsomer75TypeII(),
        ]


register_class(Sucrose_3_3 | ChaoticEntropy_3_3 | LargeWindSpirit_3_3)
